import sys
import logging
from typing import Literal, Any
from collections.abc import Callable, Iterable, Mapping
from functools import cached_property
from logging import Logger, StreamHandler, Handler
from ...misc import ArgRepr, JsonStreamHandler
from .formats import JSON_FMT

type JSON = Mapping[str, Any]


class PassThroughJsonLogger[**P](ArgRepr):
    """Pass-through Logger to stdout or stderr with a JSON-formatted messages.

    Parameters
    ----------
    name: str
        Name of the Logger. Typically set to ``__name__``.
    level: int, optional
        Minimum logging level. Defaults to 10 (= DEBUG).
    stream: str, optional
        Must be one of "stdout" or "stderr". Defaults to "stdout".
    field: str or iterable, optional
        A single `LogRecord attribute <https://docs.python.org/3/library/
        logging.html#logrecord-attributes>`_ or an iterable thereof. All values
        that are not marked by "You shouldn’t need to format this yourself."
        are allowed. Logged messages will be merged with a dictionary of these
        keys before being written to `stream` unless you explicitly refer to
        the log-record attribute "message", in which case that field will
        contain the dictionary you want to log.
        Defaults to ``("levelname", "name")``.
    *fields: str
        Additional LogRecord attributes to merge with the log message.
    **extras
        Additional, static key-value pairs to merge into every log message.

    Note
    ----
    The instantiation of the actual, underlying Logger is delayed until it is
    first needed to facilitate usage in multiprocessing scenarios.

    Warnings
    --------
    If the requested Logger already has a Handler to the requested stream, it
    will be deleted and replaced with a new one. By consequence, requesting
    the same Logger multiple times with different options might change that
    Logger globally.

    """
    def __init__(
        self,
        name: str,
        level: int = logging.DEBUG,
        stream: Literal['stdout', 'stderr'] = 'stdout',
        field: str | Iterable[str] = JSON_FMT,
        *fields: str,
        **extras: Any,
    ) -> None:
        self.name = name.strip()
        self.level = min(max(level, logging.DEBUG), logging.CRITICAL)
        self.stream = stream
        self.field = field
        self.fields = fields
        self.extras = extras
        super().__init__(
            self.name,
            self.level,
            stream,
            field,
            *fields,
            **extras
        )

    class Log:
        """Return type of the logging methods.

        Note
        ----
        This class is not intended to ever be instantiated directly.

        See Also
        --------
        PassThroughJsonLogger.debug
        PassThroughJsonLogger.info
        PassThroughJsonLogger.warning
        PassThroughJsonLogger.error
        PassThroughJsonLogger.critical

        """

        def __init__(
                self,
                parent: 'PassThroughJsonLogger',
                level: int,
                msg: JSON | Callable[[JSON], str]
        ) -> None:
            self.parent = parent
            self.level = level
            self.msg = msg

        def __call__(self, *args: P.args) -> P.args:
            """Log the cached message and pass through the input argument(s).

            Parameters
            ----------
            *args
                If `msg` is dictionary-like, the arguments are inconsequential,
                and it will simply be logged. If, however, the `msg` is
                callable, it will be called with these arguments and the
                return value will be logged instead.

            Returns
            -------
            object or tuple
                Single argument if called with only one, otherwise all of them.

            """
            msg = self.msg(*args) if callable(self.msg) else self.msg
            self.parent.logger.log(self.level, msg)
            return args[0] if len(args) == 1 else args

    def log(self, level: int, message: JSON | Callable[[P], JSON]) -> Log:
        """Log a (dictionary) message, optionally depending on call args.

        Parameters
        ----------
        level: int
            The level to log at.
        message: str or callable
            Message to log when the returned object is called. If it is
            callable, then it will be called with whatever argument(s) the
            returned object is called with and the result will be logged.

        Returns
        -------
        Log
            A callable object that logs the `message` when called.

        """
        return self.Log(self, level, message)

    def debug(self, message: JSON | Callable[[P], JSON]) -> Log:
        """Log a DEBUG (dictionary) message, optionally depending on call args.

        Parameters
        ----------
        message: str or callable
            Message to log when the returned object is called. If it is
            callable, then it will be called with whatever argument(s) the
            returned object is called with and the result will be logged.

        Returns
        -------
        Log
            A callable object that logs the `message` when called.

        """
        return self.Log(self, logging.DEBUG, message)

    def info(self, message: JSON | Callable[[P], JSON]) -> Log:
        """Log an INFO (dictionary) message, optionally depending on call args.

        Parameters
        ----------
        message: Mapping or callable
            (Dictionary) message to log when the returned object is called.
            If it is callable, then it will be called with whatever argument(s)
            the returned object is called with and the result will be logged.

        Returns
        -------
        Log
            A callable object that logs the `message` when called.

        """
        return self.Log(self, logging.INFO, message)

    def warning(self, message: JSON | Callable[[P], JSON]) -> Log:
        """Log a WARNING (dict) message, optionally depending on call args.

        Parameters
        ----------
        message: Mapping or callable
            (Dictionary) message to log when the returned object is called.
            If it is callable, then it will be called with whatever argument(s)
            the returned object is called with and the result will be logged.

        Returns
        -------
        Log
            A callable object that logs the `message` when called.

        """
        return self.Log(self, logging.WARNING, message)

    def error(self, message: JSON | Callable[[P], JSON]) -> Log:
        """Log an ERROR (dict-like) message, optionally depending on call args.

        Parameters
        ----------
        message: Mapping or callable
            (Dictionary) message to log when the returned object is called.
            If it is callable, then it will be called with whatever argument(s)
            the returned object is called with and the result will be logged.

        Returns
        -------
        Log
            A callable object that logs the `message` when called.

        """
        return self.Log(self, logging.ERROR, message)

    def critical(self, message: JSON | Callable[[P], JSON]) -> Log:
        """Log a CRITICAL (dict) message, optionally depending on call args.

        Parameters
        ----------
        message: Mapping or callable
            (Dictionary) message to log when the returned object is called.
            If it is callable, then it will be called with whatever argument(s)
            the returned object is called with and the result will be logged.

        Returns
        -------
        Log
            A callable object that logs the `message` when called.

        """
        return self.Log(self, logging.CRITICAL, message)

    @cached_property
    def logger(self) -> Logger:
        """The requested Logger with exactly one JsonStreamHandler."""
        # Get the Logger with the given name
        logger = logging.getLogger(self.name)
        # Adjust its log level so that messages from the Handler get through
        logger.setLevel(self.level)
        # Remove all existing Handlers to the requested stream
        for handler in self.__filtered(logger.handlers):
            logger.removeHandler(handler)
        # Create a new JsonStreamHandler
        handler = JsonStreamHandler(
            self.stream,
            self.field,
            *self.fields,
            **self.extras
        )
        handler.setLevel(self.level)
        # ... add the new one instead, ...
        logger.addHandler(handler)
        # ... and return the logger.
        return logger

    def __filtered(self, handlers: list[Handler]) -> tuple[StreamHandler, ...]:
        """Filter the Logger's Handlers according to the filter criterion."""
        return tuple(filter(self.__handles_the_same_stream, handlers))

    def __handles_the_same_stream(self, handler: Handler) -> bool:
        """Is the Handler a StreamHandler to the requested stream?"""
        if isinstance(handler, StreamHandler):
            return handler.stream is getattr(sys, self.stream)
        return False



import sys
import json
import logging
from typing import Any, Literal
from collections.abc import Iterable, Mapping
from functools import cached_property
from datetime import datetime
from logging import Logger, StreamHandler, Handler, LogRecord
from ..repr import ArgRepr
from .formats import JSON_FMT


class JsonStreamHandler(StreamHandler):
    """StreamHandler for logging JSON-formatted messages.

    Parameters
    ----------
    stream: str, optional
        Must be one of "stdout" or "stderr". Defaults to "stdout".
    field: str or iterable, optional
        A single `LogRecord attribute <https://docs.python.org/3/library/
        logging.html#logrecord-attributes>`_ or an iterable thereof. All values
        that are not marked by "You shouldn’t need to format this yourself."
        are allowed. Logged messages will be merged with a dictionary of these
        keys before being written to `stream` unless you explicitly refer to
        the LogRecord attribute "message", in which case that field will
        contain the dictionary you want to log.
        Defaults to ``("levelname", "name")``.
    *fields: str
        Additional LogRecord attributes to merge with the log message.
    **extras
        Additional, static key-value pairs to merge into every log message.

    Raises
    ------
    TypeError
        If `stream` is not a string.
    ValueError
        If `stream` is neither "stdout" nor "stderr" or if any of the given
        `fields` are not a permissible LogRecord attribute.

    """

    __whitelist__ = {
        'asctime',
        'created',
        'filename',
        'funcName',
        'levelname',
        'levelno',
        'lineno',
        'message',
        'module',
        'msecs',
        'name',
        'pathname',
        'process',
        'processName',
        'relativeCreated',
        'thread',
        'threadName',
        'taskName'
    }

    def __init__(
            self,
            stream: Literal['stdout', 'stderr'] = 'stdout',
            field: str | Iterable[str] = JSON_FMT,
            *fields: str,
            **extras: Any
    ) -> None:
        super().__init__(stream=getattr(sys, self.__valid(stream)))
        self.fields = self.__permissible(field) | self.__permissible(fields)
        self.extras = extras

    @staticmethod
    def __valid(stream: str) -> str:
        """Ensure that the provided stream is one of the permitted options."""
        if not isinstance(stream, str):
            cls = type(stream).__name__
            msg = f'stream must be a string, not {cls}!'
            raise TypeError(msg)
        stream = stream.strip().lower()
        if stream not in ('stdout', 'stderr'):
            msg = f'stream must be "stdout" or "stderr", not "{stream}"!'
            raise ValueError(msg)
        return stream

    def __permissible(self, fields: str | Iterable[str]) -> set[str]:
        """Ensure that fields are white-listed strings or iterables thereof."""
        if isinstance(fields, str):
            return {self.__whitelisted(fields.strip().lower())}
        return {self.__whitelisted(str(f).strip().lower()) for f in fields}

    def __whitelisted(self, field: str) -> str:
        """Ensure that every requested field is on the whitelist."""
        if field in self.__whitelist__:
            return field
        else:
            tmp = '"{}" is not a valid log-record attribute'
            raise ValueError(tmp.format(field))

    @property
    def basics(self) -> set[str]:
        """LogRecord attributes to add to the log message."""
        return self.fields - {'message'}

    def format(self, record: LogRecord) -> str:
        """Format the LogRecord as a JSON-serialized string."""
        # To simplify, add converted timestamp, even if we don't need it
        record.asctime = self.asctime(record.created)

        # Dictionary with requested fields plucked from the log record
        basics = {basic: getattr(record, basic) for basic in self.basics}

        # If requested, add the log message as a nested field ...
        if 'message' in self.fields:
            message = basics | {'message': record.msg}
        # ... or, if not, merge the requested fields with the message.
        else:
            # If the message is indeed a dictionary, this should work ...
            try:
                message = record.msg | basics
            # ... but if it does not, we fall back to a nested message.
            except TypeError:
                message = basics | {'message': record.msg}

        # Return the serialized JSON as the log message to be emitted.
        return json.dumps(
            self.extras | message,
            sort_keys=True,
            default=lambda x: x.as_json if hasattr(x, 'as_json') else repr(x),
        )

    @staticmethod
    def asctime(timestamp: float) -> str:
        """Format a unix timestamp to string, including milliseconds."""
        return datetime.fromtimestamp(timestamp).strftime(
            '%Y-%m-%d %H:%M:%S.%f'
        )[:-3]


class JsonLogger(ArgRepr):
    """Wrapped logger to stdout or stderr with a JSON-formatted StreamHandler.

    Parameters
    ----------
    name: str
        Name of the Logger. Typically set to ``__name__``.
    level: int, optional
        Minimum logging level. Defaults to 10 (= DEBUG).
    stream: str, optional
        Must be one of "stdout" or "stderr". Defaults to "stdout".
    field: str or iterable, optional
        A single `LogRecord attribute <https://docs.python.org/3/library/
        logging.html#logrecord-attributes>`_ or an iterable thereof. All values
        that are not marked by "You shouldn’t need to format this yourself."
        are allowed. Logged messages will be merged with a dictionary of these
        keys before being written to `stream` unless you explicitly refer to
        the LogRecord attribute "message", in which case that field will
        contain the dictionary you want to log.
        Defaults to ``("levelname", "name")``.
    *fields: str
        Additional LogRecord attributes to merge with the log message.
    **extras
        Additional, static key-value pairs to merge into every log message.

    Note
    ----
    The instantiation of the actual, underlying Logger is delayed until it is
    first needed to facilitate usage in multiprocessing scenarios.

    Warnings
    --------
    If the requested Logger already has a Handler to the requested stream, it
    will be deleted and replaced with a new one. By consequence, requesting
    the same Logger multiple times with different options might change that
    Logger globally.

    """
    def __init__(
            self,
            name: str,
            level: int = logging.DEBUG,
            stream: Literal['stdout', 'stderr'] = 'stdout',
            field: str | Iterable[str] = JSON_FMT,
            *fields: str,
            **extras: Any,
    ) -> None:
        self.name = name.strip()
        self.level = min(max(level, logging.DEBUG), logging.CRITICAL)
        self.stream = stream
        self.field = field
        self.fields = fields
        self.extras = extras
        super().__init__(
            self.name,
            self.level,
            stream,
            field,
            *fields,
            **extras
        )

    def log(self, level: int, message: Mapping[str, Any]) -> tuple[()]:
        """Log a message at the given level.

        Parameters
        ----------
        level: int
            The level to log at.
        message: Mapping
            The (dictionary) message to log.

        Returns
        -------
        tuple
            An empty tuple.

        """
        self.logger.log(level, message)
        return ()

    def debug(self, message: Mapping[str, Any]) -> tuple[()]:
        """Log a DEBUG message.

        Parameters
        ----------
        message: Mapping
            The (dictionary) message to log.

        Returns
        -------
        tuple
            An empty tuple.

        """
        return self.log(logging.DEBUG, message)

    def info(self, message: Mapping[str, Any]) -> tuple[()]:
        """Log an INFO message.

        Parameters
        ----------
        message: Mapping
            The (dictionary) message to log.

        Returns
        -------
        tuple
            An empty tuple.

        """
        return self.log(logging.INFO, message)

    def warning(self, message: Mapping[str, Any]) -> tuple[()]:
        """Log a WARNING message.

        Parameters
        ----------
        message: Mapping
            The (dictionary) message to log.

        Returns
        -------
        tuple
            An empty tuple.

        """
        return self.log(logging.WARNING, message)

    def error(self, message: Mapping[str, Any]) -> tuple[()]:
        """Log an ERROR message.

        Parameters
        ----------
        message: Mapping
            The (dictionary) message to log.

        Returns
        -------
        tuple
            An empty tuple.

        """
        return self.log(logging.ERROR, message)

    def critical(self, message: Mapping[str, Any]) -> tuple[()]:
        """Log a CRITICAL message.

        Parameters
        ----------
        message: Mapping
            The (dictionary) message to log.

        Returns
        -------
        tuple
            An empty tuple.

        """
        return self.log(logging.CRITICAL, message)

    @cached_property
    def logger(self) -> Logger:
        """The requested Logger with exactly one JsonStreamHandler."""
        # Get the Logger with the given name
        logger = logging.getLogger(self.name)
        # Adjust its log level so that messages from the Handler get through
        logger.setLevel(self.level)
        # Remove all existing Handlers to the requested stream
        for handler in self.__filtered(logger.handlers):
            logger.removeHandler(handler)
        # Create a new JsonStreamHandler
        handler = JsonStreamHandler(
            self.stream,
            self.field,
            *self.fields,
            **self.extras
        )
        handler.setLevel(self.level)
        # ... add the new one instead, ...
        logger.addHandler(handler)
        # ... and return the logger.
        return logger

    def __filtered(self, handlers: list[Handler]) -> tuple[StreamHandler, ...]:
        """Filter the Logger's Handlers according to the filter criterion."""
        return tuple(filter(self.__handles_the_same_stream, handlers))

    def __handles_the_same_stream(self, handler: Handler) -> bool:
        """Is the Handler a StreamHandler to the requested stream?"""
        if isinstance(handler, StreamHandler):
            return handler.stream is getattr(sys, self.stream)
        return False



from typing import Any, Literal
from collections.abc import Callable
from numpy import ndarray
from pandas import DataFrame, Series
from pandas.core.window import RollingGroupby
from pandas.core.groupby import DataFrameGroupBy, SeriesGroupBy
from ..misc import ArgRepr

type GroupBy = SeriesGroupBy | DataFrameGroupBy
type Pandas = Series | DataFrame


class RollingGroupByApply(ArgRepr):
    """Partial for calling a rolling-grouped dataframe's ``apply`` method.

    Parameters
    ----------
    func: callable
        Must produce a single numerical value from a numpy ndarray if
        `raw` = ``True`` or a series if `raw` = ``False``.
        Can also accept a numba JIT function with `engine` = "numba" specified.
    raw: bool, optional
        Whether to pass a numpy ndarray or a pandas series to `func`.
        Defaults to ``False``.
    engine: str, optional
        Either "cython" or "numba". Defaults to ``None``
    engine_kws: dict, optional
        Configuration of the "numba" engine. Keys can be "nopython", "nogil",
        and "parallel", and values must be ``True`` or ``False``.
        Defaults to ``None``.
    *args
        Positional arguments to pass to `func`.
    **kwargs
         Keyword arguments to pass to `func`.

    """

    def __init__(
            self,
            func: Callable[[Series | ndarray], float],
            raw: bool = False,
            engine: Literal['cython', 'numba'] | None = None,
            engine_kws: dict[str, bool] | None = None,
            *args: Any,
            **kwargs: Any
    ) -> None:
        super().__init__(func, raw, engine, engine_kws, *args, **kwargs)
        self.func = func
        self.raw = raw
        self.engine = engine
        self.engine_kws = engine_kws
        self.args = args
        self.kwargs = kwargs

    def __call__(self, rolling_df: RollingGroupby) -> DataFrame:
        """Call a rolling-grouped dataframe`s ``apply`` method.

        Parameters
        ----------
        rolling_df: RollingGroupby
            The rolling-grouped dataframe to aggregate.

        Returns
        -------
        DataFrame
            The aggregation of the rolling-grouped dataframe.

        """
        return rolling_df.apply(
            self.func,
            self.raw,
            self.engine,
            self.engine_kws,
            self.args,
            self.kwargs
        )


class GroupByApply(ArgRepr):
    """Partial for calling a grouped dataframe or series ``apply`` method.

    Parameters
    ----------
    func: callable
        A callable that takes a dataframe or series as its first argument,
        and returns a dataframe, a series or a scalar.
    *args
        Positional arguments to pass to `func`.
    **kwargs
         Keyword arguments to pass to `func`.

    """

    def __init__(
            self,
            func: Callable[[Pandas, ...], Any],
            *args: Any,
            **kwargs: Any
    ) -> None:
        super().__init__(func, *args, **kwargs)
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def __call__(self, df: GroupBy) -> Pandas:
        """Call a grouped dataframe or series ``apply`` method.

        Parameters
        ----------
        df: DataFrameGroupBy or SeriesGroupBy
            The pandas group-by object to apply `func` to.

        Returns
        -------
        Series or DataFrame
            The input object with `func` applied to groups.

        """
        return df.apply(self.func, *self.args, **self.kwargs)



from .formats import DEFAULT_FMT, PID_FMT, RAW_FMT, SHORT_FMT, JSON_FMT
from .std import StdLogger
from .file import FileLogger
from .json import JsonLogger, JsonStreamHandler

__all__ = [
    'DEFAULT_FMT',
    'SHORT_FMT',
    'PID_FMT',
    'RAW_FMT',
    'JSON_FMT',
    'StdLogger',
    'FileLogger',
    'JsonLogger',
    'JsonStreamHandler'
]
